#include <pic.h>

	/* konfiguracja procesora */							/* 0x3FF5 / 0x3FF1*/
__CONFIG(UNPROTECT&PWRTEN&WDTEN&HS); /* NO_PROTECT, PWRT_EN, WDT_DIS, HS */

						/* SERIAL NUMBER procesora */
__EEPROM_DATA(0x53, 0x41, 0x57, 0x49, 0xA9, 0x4B, 0x49, 0xFF);

//===========================================================================

#define byte unsigned char
#define word unsigned int

#define	XTAL_FREQ	10MHz			/* Crystal frequency in MHz */
#define	MHz	*1000L				/* number of kHz in a MHz */
#define	Delay_us(x)	  { byte _dcnt; _dcnt = (x)/((12MHz)/(XTAL_FREQ))|1; \
		  						 while(--_dcnt != 0) continue; }

#define TRUE	1
#define FALSE	0

#define FULL	1
#define EMPTY	0

#define INPUT  1
#define OUTPUT 0

#define M_R_C				0x01		/* Master ROM Command */
#define M_F_C				0x02    	/* Master Function Command */

											/* Init Sequence - sekwencja Reset Pulse */
#define INIT_SEQ		  ( flaga_Tx=0, flaga_Rx=0, flaga_BUF_Rx=0, \
							    flaga_busy=0, flaga_Match_ROM=0, licznik_bit=0, \
								 status=0x00 )

//------------------- definicje pinow we / wy -------------------------------

#define MASKA_RB	0x01  /* obowiazkowo RB0 jako INPUT */

						/* obsuga magistrali 1-wire */
static bit DQ 					@ ((unsigned)&PORTB*8+0);
static bit DQ_TRIS			@ ((unsigned)&TRISB*8+0);

//-------------------------- zmienne ogolne ---------------------------------

bit 	flaga_Rx, flaga_Tx;			// flagi trybu pracy 1-wire (nadawanie/odbior)
bit	flaga_BUF_Rx;					// odebrano 8 bitow od 'Mastera'
bit	flaga_Match_ROM;
bit 	flaga_busy;						// konwwersja np. temp i wilgotn w trakcie !
byte	status;							// aktualna sekwencja w ROM Commands Flow
											// i w Function Commands Flow

byte 	BUF_Rx;							// bufor do odbierania komend od 'Mastera'
bit	bRx;								// bit odebranej transmisji 1-wire
byte 	BUF_Tx[8];						// bufor nadawania danych do 'Mastera'
byte	max_index_buf;					// ilosc nadawanych bytow danych
byte 	SN[8];							// bufor z Serial Number

byte 	licznik_bit,					// licznik wysylanych bit'ow
		index_buf,						// index aktualnie wysylanego bytu bufora Tx
		del_count;						// licznik do odmierzania opoznienia

//============================== procedury =================================

void interrupt ISR(void);	// procedura odbioru i transmisji 1-wire
void Delay_ms(byte cnt);	// generowanie opoznienia w milisekundach
void Read_EE(byte adres);	// odczyt danych z wewn. eepromu
void C_CRC(byte *CRCVal, byte value);  	// generator CRC
byte CalcCRC(byte code_len, byte *code);   // generator CRC

//===========================================================================
void interrupt ISR(void)
//---------------------------------------------------------------------------
{
/*===WRITE BIT===	a) sprawdzamy czy jest w trybie nadawania 'flaga_Tx' -
							jezeli tak to wystawiamy DQ w zaleznosci od danych w
							buforze Tx przez max 15us
						b) odczekujemy 20us (przygotowanie bufora do nastepnej
							transmisji)
						c) oczekujemy przez dalsze 20us, az DQ==1 i opuszczamy
							procedure, gotowi	do transmisji nastepnego bitu
						d) jezeli po tym czasie DQ != 1 to znaczy	ze prawdopodobnie
							jest to (niespodziewany) sygnal Reset i idziemy do etykiety  RST */
	if (flaga_Tx)
   {	// ta sekwencja nie moze byc zmieniana ze wzgledu na rygor czasow
/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
/*!!*/if (BUF_Tx[index_buf] & 0x01)
/*!!*/{//---transmisja poszczegolnych bitow z bufora nadawczego
/*!!*/	DQ_TRIS = INPUT;	// DQ = 1;
/*!!*/}
/*!!*/else
/*!!*/{
/*!!*/   DQ_TRIS = OUTPUT;
/*!!*/	DQ = 0;
/*!!*/   Delay_us(10);		// @@@ oryginalna wartosc 10
/*!!*/	DQ_TRIS = INPUT;	// DQ = 1;
/*!!*/}
/*!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/
      BUF_Tx[index_buf] >>= 1;	// przygotowanie kolejnego bitu do nadawania

      if (++licznik_bit == 8)
      {
         if (flaga_busy)     		// jako znacznik konca procesu np. konwersja
         	index_buf = 0;			// temperatury -> nadawany jest non-stop BUF_Tx[0]
         else
            if (index_buf++ == max_index_buf)
					INIT_SEQ;

         licznik_bit = 0;
	   }
      //---kontrola poprawnosci transmisji
      if (DQ)
      {
         INTF = 0;
			return;
      }
      else
      {
         del_count = 45;			// opoznienie ~= Delay_us(45)
      	while (del_count--)
	      {
				if (DQ)
      		{
               INTF = 0;
					return;
            }
         }
			goto RST;
		}
	}
//---------------------------------------------------------------------------
/*===READ BIT===	a) czy jest gotowy do odbioru 'flaga_Rx' (tzn. ze byl
							poprawny Reset)
						b) odczekujemy min 15us i sprawdzamy DQ :
							Jezeli DQ==0 i stan ten trwa 60<T<120us a potem DQ==1
								to znaczy, ze odebralismy '0' i opuszczamy procedure.
         				Jezeli DQ==1 to znaczy, ze odebralismy '1' i opuszczamy
								procedure
      					Jezeli DQ==0 ponad 120us -> moze to oznaczac niespodziewany
								Reset	i nalezy isc do etykiety RST, w ktorej zostanie
								sprawdzona poprawnosc Reset Pulse */
   if (flaga_Rx)
   {
      //---odczyt i zapis danych do bufora odbiorczego
      Delay_us(15);		// @@@ 15
      bRx = DQ;
      //---kontrola ilosci odebranych bitow
      flaga_BUF_Rx = EMPTY;			// znacznik odebrania ostatniego 8 bitu
      if (++licznik_bit == 8)
      {
     		flaga_BUF_Rx = FULL;		// odebrano 8 ostatni bit
         index_buf++;
         licznik_bit = 0;
      }
      BUF_Rx >>= 1;
      if (bRx)
      	BUF_Rx |= 0x80;
		Delay_us(5);
      //---sprawdzenie poprawnosci procedury odbioru bitu
      if (DQ)
      {
         INTF = 0;					// reset flagi przerwania
			return;						// opuszczenie proc. obslugi przerwania
      }
      else
      {
         del_count = 50;			//* @@@ 50
      	while (del_count--)		// opoznienie ~= Delay_us()
	      {
				if (DQ)
      		{
               INTF = 0;			// reset flagi przerwania
					return;				// opuszczenie proc. obslugi przerwania
            }
         }
			// goto RST;				// naturalnie przechodzi do RST
		}
	}
//---------------------------------------------------------------------------
	RST :
/*--- Procedura identyfikacji Reset'u oraz wysylania 'Presence Pulse' ---*/
	/*--- odczekujemy pozostaly czas do ok. 480 us ---*/
   /* pierwotnie 250 + 70 -> 250 */
	Delay_us(200); // Delay_us(50);		/* bez ww. procedur us=250+150 */
						/* jezeli czas opoznienia powyzej jest za dlugi -
							250 i 70 to pierwszy RESET jest OK ale nastepne ERROR,
							odczyt SN jest poprawny przy kazdym wariancie */
   /*--- czekamy az bedzie stan HI */
	while(DQ == 0);
   Delay_us(10);
   /*--- wystawiamy sygnal 'Presence Pulse' */
   DQ_TRIS = OUTPUT;
	DQ = 0;

   /*--- po ok. 200us przestawiamy port na INPUT */
 	Delay_us(200);
   {
      /*--- kasujemy bufor_Rx i flagi i liczniki -> gotowy do odbioru */
      INIT_SEQ;
      flaga_Rx = TRUE;
      status 	= M_R_C;
   }

   DQ_TRIS = INPUT;	//	DQ = 1;

   INTF = 0;
   return;
}

//---------------------------------------------------------------------------
void Delay_ms(byte cnt)
//---------------------------------------------------------------------------
{
	byte i;

	do
	{
		i = 4;
		do
		{
			Delay_us(206);	//195
		}
		while(--i);
	}
	while(--cnt);
}

//---------------------------------------------------------------------------
void Read_SN(void)
//---------------------------------------------------------------------------
{
   byte adres = 0x00;

   do
   {
		EEADR = adres;
   	RD = 1;
   	SN[adres] = EEDATA;
   }while(++adres < 7);		// odczytujemy tylko 7 bytow, 8 byte jako CRC jest
   								// wyliczany w ponizszej procedurze !
	SN[adres] = CalcCRC(7, SN);
}
//%%%%%%%%%%%%%%%%%% generator Cyclic Redundancy Checks %%%%%%%%%%%%%%%%%%%%%

//---------------------------------------------------------------------------
void C_CRC(byte *CRCVal, byte value)
//---------------------------------------------------------------------------
{
  	byte odd, bcnt;

  	for (bcnt=0; bcnt<8; bcnt++)
	{
		odd = (value ^ *CRCVal) & 0x01;
	 	*CRCVal >>= 1;
	 	value >>= 1;
	 	if (odd != 0)
			*CRCVal ^= 0x8C;
  	}
}

//---------------------------------------------------------------------------
byte CalcCRC(byte code_len, byte *code)
//---------------------------------------------------------------------------
{
	byte i, CRCVal=0;

  	for ( i=0; i<code_len; i++ )
    	C_CRC( &CRCVal, code[i] );

  	return CRCVal;
}

//***************************************************************************
//************************* 			MAIN		********************************
//***************************************************************************
void main(void)
{
	byte i;

	/* konfiguracja portow RA i RB [ RB0 muszi byc jako INPUT */
	TRISA = 0xFF;						// xxxx xxxx
 	TRISB = 0xFF | MASKA_RB;		// xxxx xxx1

   INIT_SEQ;							// sekwencja resetujaca ustawienia proc. INT

   Read_SN();							// ustalenie Serial Number dla ukladu

   BUF_Rx = 0;							// wyzerowanie bufora odbiorczego

	/* konfiguracja przerwan */
   INTEDG = 0;							// przerwanie na RB0 opadajacym zboczem
   INTE = 1;							// zezwolanie na przerwania na Ext_INT (RB0)

   GIE = 1;								// Global Interrupt Enable

	//------------------------------------------------------------------------

 	while(1)
	{
      if (flaga_BUF_Rx == FULL)			// buforze Rx jest 8 bitow danych
      {
//%%%%%%%%%%%%%%%%%%%%%%%%%% MASTER ROM COMMAND %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
         if (status == M_R_C)
         {
            if (flaga_Match_ROM)			// kontrola Match_ROM
            {
         	   if (BUF_Rx != SN[i++])	// kontrola poprawnosci SN
                  INIT_SEQ;				// powrot do poczatku Reset Pulse

               if (i == 8)					// sprawdzono wszystki 8 bytow i sa OK
                  status = M_F_C;		// status na Master Function Commands

               flaga_BUF_Rx = EMPTY;	// czyszczenie bufora na przyjecie kolejnego bytu
               goto END_MRC;				// aby opuscil procedure nie idac ponizej
			   }

            switch (BUF_Rx)
            {
/* Read ROM */ case 0x33 :	// -> transmisja Family Code (Serial Number)
                  //---zakonczenie odbioru
                  flaga_Rx = FALSE;
                  flaga_BUF_Rx = EMPTY;
                  //---przepisanie Serial Number do bufora nadawczego
                  for(i=0; i<8; i++)
                     BUF_Tx[i] = SN[i];
                  //---ustalenie rozmiarow bufora nadawczego
                  index_buf = 0;
                  max_index_buf = 7;	// 8-bytowy buf. nadawczy
                  //---przestawienie sie na nadawanie
                  flaga_Tx = TRUE;
               	break;
/* Match ROM */case 0x55 :	// -> odbior Family Code (Serial Number)
                  //---zakonczenie odbioru bytu 0x55 -> oczekiwanie na kolejne
                  flaga_BUF_Rx = EMPTY;
                  //---przejscie do procedury sprawdzania Serial Number
					   flaga_Match_ROM = TRUE;
                  //---zerujemy licznik odbieranych bytow danych np. SN
                  i = 0;
               	break;
/* Skip ROM */ case 0xCC : // -> przejscie do trybu Master Function Command
						//---zakonczenie odbioru bytu 0xCC -> oczekiwanie na kolejne
                  flaga_BUF_Rx = EMPTY;
                  //---zmiana statusu na Master Function Commands
		            status = M_F_C;
               	break;
		      }
            END_MRC: CLRWDT();
         }
//%%%%%%%%%%%%%%%%%%%%%%%% MASTER FUNCTION COMMAND %%%%%%%%%%%%%%%%%%%%%%%%%%
         else if ((status & 0x0F) == M_F_C)
         {
            switch (status)
            {
            	case 0x42 :
               	PORTA = BUF_Rx;
                  INIT_SEQ;
                  goto END_MFC;
            	case 0x52 :
               	PORTB = BUF_Rx | MASKA_RB;
                  INIT_SEQ;
                  goto END_MFC;
            	case 0x72 :
               	TRISA = BUF_Rx;
                  INIT_SEQ;
                  goto END_MFC;
            	case 0x82 :
               	TRISB = BUF_Rx | MASKA_RB;
                  INIT_SEQ;
                  goto END_MFC;
				}

         	switch (BUF_Rx)
            {
            	case 0xF4 :					// WRITE to RA
                  status = 0x42;
                  flaga_BUF_Rx = EMPTY;
                  break;
            	case 0xF5 :					// WRITE to RB
                  status = 0x52;
                  flaga_BUF_Rx = EMPTY;
                  break;
            	case 0xF6 :					// READ from RA & RB
                  //---koniec odbioru
                  flaga_Rx = FALSE;
                  flaga_BUF_Rx = EMPTY;
                  //---przygotowanie bufora nadawczego
                  BUF_Tx[0] = PORTA;
                  BUF_Tx[1] = PORTB;
                  index_buf = 0;
                  max_index_buf = 1;
                  //---przelaczenie sie na nadawanie
                  flaga_Tx = TRUE;
                  break;
            	case 0xF7 :					// SETUP PORT RA [ - - - x x x x x ]
                  status = 0x72;
                  flaga_BUF_Rx = EMPTY;
                  break;
            	case 0xF8 :					// SETUP PORT RB [ x x x x x x x DQ]
                  status = 0x82;
                  flaga_BUF_Rx = EMPTY;
                  break;
				}
            END_MFC: CLRWDT();
			}
      }
   	CLRWDT();
   }
}
